/*
 * Copyright (c) 2002, RISS-Telecom. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 4. Neither the name of the RISS-Telecom nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * The purpose of these functions is to test trafd behaviour
 * with different hash table functions.
 * hash_tests() function mimics trafd in process of collecting
 * traffic records while measuring hash functions performance.
 * Although this test is not perfect it clearly shows that
 * trafd performs 30-40 times faster with new (XOR) hash function
 * on large data tables.
 */

#ifdef WITH_HASH_TESTS

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "traffic.h"
#include "util.h"

extern struct t_entry *entries;
extern int n_entry;

int64_t hash_chain_len, hash_chain_n;
int32_t hash_chain_max;

void
hash_tests_insertdata(p_entry pte)
{
	int t_idx;

	/* here we call hash function */
	t_idx = insertentry(pte);

	if (n_entry > MAX_TO_SAVE || t_idx == -2) {
		fprintf(stderr, "MAX_TO_SAVE=%d reached, clearing table!\n",
			MAX_TO_SAVE);
		memset((struct t_entry *)entries, 0,
			sizeof(struct t_entry) * MAX_T_ENTRIES);
		n_entry = 0;
        return;
	}
	if (t_idx >= 0)
		entries[t_idx].n_bytes += pte->n_bytes,
		entries[t_idx].n_psize += pte->n_psize;
	else
		n_entry++;
	return;
}

/* test hash function, read trafd data file from stdin into a hash */
void
hash_tests(void)
{
	struct t_header th;
	struct t_entry te;
	int i;
	struct timeval tvs, tv;
	struct rusage ru;
	long nall = 0;

	n_entry = 0;
	entries = (struct t_entry *) calloc(MAX_T_ENTRIES, sizeof(struct t_entry));
	if (entries == NULL) {
		perror("calloc");
		exit(1);
	}

	gettimeofday(&tvs, NULL);

	while (read(0, &th, sizeof(th)) > 0) {
	    if (th.t_size > MAX_TO_SAVE)
			error("table too big to fit in memory");
		if (!th.t_size)
			continue;
	
		for (i = 0; i < th.t_size; i++) {
			if (read(0, &te, sizeof(te)) != sizeof(te))
				error("can't read entry");
			hash_tests_insertdata(&te);
		}

		fprintf(stderr, "%d ins, %d in table, %d max, %d to save\n", th.t_size, n_entry, MAX_T_ENTRIES, MAX_TO_SAVE);
		nall += th.t_size;
	}

	getrusage(RUSAGE_SELF, &ru);
	gettimeofday(&tv, NULL);
	tv.tv_sec -= tvs.tv_sec;
	tv.tv_usec -= tvs.tv_usec;
	if (tv.tv_usec < 0)
		tv.tv_sec--, tv.tv_usec += 1000000;

	fprintf(stderr, "-- \n");
	fprintf(stderr, "processed %ld, %d elem in table\n", nall, n_entry);
	fprintf(stderr, "times: %ld.%03ld real, %ld.%03ld user, %ld.%03ld system\n",
		tv.tv_sec, tv.tv_usec/1000,
		ru.ru_utime.tv_sec, ru.ru_utime.tv_usec/1000,
		ru.ru_stime.tv_sec, ru.ru_stime.tv_usec/1000);
	fprintf(stderr, "MAX_TO_SAVE=%d, MAX_T_ENTRIES=%d\n", MAX_TO_SAVE, MAX_T_ENTRIES);
	fprintf(stderr, "avg hash chain len %.3f, maximum %d\n", ((double)hash_chain_len)/hash_chain_n, hash_chain_max);
}

#endif /* WITH_HASH_TESTS */
