/*
 *	Copyright (c) 1999-2002 RISS-Telecom Networking Center
 *
 *	Copyright (c) 1993 The CAD lab of the
 *	Novosibirsk Institute of Broadcasting and Telecommunication
 *
 *	BPFT $Id: main.c,v 1.3 2006/09/27 03:13:37 count Exp $
 *
 *	$Log: main.c,v $
 *	Revision 1.3  2006/09/27 03:13:37  count
 *	Fix NOMAN to NO_MAN
 *	s/savestr/strdup/
 *	Some prepare for global libtrafdb.
 *	
 *	Revision 1.1.1.1  2000/04/07 05:26:48  count
 *	Initial revision
 *	
 * Revision 1.4  1994/01/14  11:46:41  bob
 * Any file extention maybe used, extra needed for FreeBSD, thanx vv@sonet.kemerovo.su
 *
 * Revision 1.3  1994/01/06  14:59:17  bob
 * Cosmetic changes
 *
 * Revision 1.2  1993/11/01  14:57:07  bob
 * Improve signal handlers, added SIGALRM and SIGCHLD
 *
 * Revision 1.1  1993/10/20  16:04:41  bob
 * Initial revision
 *
 *
 * Redistribution and use in source forms, with and without modification,
 * are permitted provided that this entire comment appears intact.
 * Redistribution in binary form may occur without any restrictions.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' WITHOUT ANY WARRANTIES OF ANY KIND.
 */

/*	main.c - tcp/udp data traffic collector daemon	*/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <syslog.h>
#include <paths.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/resource.h>
#include <sys/un.h>
#include <limits.h>
#include <netinet/in.h>
#include <pcap.h>

#include "interface.h"
#include "pathnames.h"
#include "addrtoname.h"
#include "util.h"
#include "traffic.h"

#include "trafd.h"

struct trafd_stats {
	struct timeval tv;
	struct pcap_stat ps;
	struct rusage ru;
};

/*
 * Command line switches.
 * Left - default value, right - action, if switch has been used.
 */
int dflag;	/* dump compiled packet-matching code to stdout and stop */
int Oflag = 1;	/* don't run the packet-matching code optimizer if have bug */
int pflag;	/* don't put the interface into promiscuous mode */
int rflag;	/* attempt to resume data from save file if exist */
int nflag;
int Nflag;
int fflag;
int Cflag = 0;	/* chroot to TRAFDDIR */

/* global signal flags */
int sigflag_save;		/* SIGUSR1 - append collected data to savefile */
int sigflag_backup;		/* SIGUSR2 - drop collected data to backup file for recovery */
int sigflag_shutdown;	/* dump data for recovery and exit */
#if 0
int sigflag_info;		/* log statistics */
#endif

/* Global variables */
char *device_name = 0;	/* interface name */
char file_pid[PATH_MAX];	/* pid file */
char file_out[PATH_MAX];	/* temp file */
char file_backup[PATH_MAX];	/* save file */
char file_sock[PATH_MAX];	/* communication socket */
char *user = NULL;
char *group = NULL;
int droproot = 0;
int daemonize = 1;
u_int save_min_record_size = MIN_RECORD_SIZE,
	pipe_min_record_size = MIN_RECORD_SIZE;

static void cleanup_exit();
static int main_loop(pcap_t*, int, int);
/* signal handlers */
static void sighandler_shutdown();	/* backup and shutdown cleanly */
static void sighandler_save();		/* save */
static void sighandler_backup();	/* backup */
static void sighandler_child();		/* reap all children */
#if 0
static void sighandler_info();		/* request info */
#endif
/* Trafd internal statistics routines */
static inline int init_trafd_stats(void);
static int get_trafd_stats(char *, pcap_t *,
	struct trafd_stats **, struct trafd_stats **);
static int log_trafd_stats(char *, pcap_t *);

/* Length of saved portion of packet */
int snaplen = DEFAULT_SNAPLEN;

static pcap_t *pd = NULL;

#ifdef WITH_HASH_TESTS
void hash_tests();
#endif

int
main(argc, argv)
	int argc;
	char **argv;
{
	void usage();

	int op, cnt = -1, i;
	struct bpf_program fcode;
	char *infile = 0, *cmdbuf, *DevFileName = 0;
	char ebuf[PCAP_ERRBUF_SIZE];
	extern char *optarg;
	extern int optind, opterr;
	bpf_u_int32 localnet, netmask;
	struct sockaddr_un sun;
	int comm_fd;
	uid_t uid = 0;
	gid_t gid = 0;

#ifdef WITH_HASH_TESTS
	if (getenv("HASH_TESTS")) {
		hash_tests();
		exit(0);
	}
#endif
   
	set_program_name(argv[0]);

	opterr = 0;
	while ((op = getopt(argc, argv, "c:Cdf:F:i:Opru:g:Dm:M:")) != EOF)
		switch (op) {
		case 'c':
			cnt = atoi(optarg);
			break;
		case 'C':
			Cflag++;
			break;
		case 'd':
			++dflag;
			break;
		case 'f':
			DevFileName = optarg;
			break;
		case 'F':
			infile = optarg;
			break;
		case 'i':
			device_name = optarg;
			break;
		case 'O':
			Oflag = 0;
			break;
		case 'p':
			++pflag;
			break;
		case 'r':
			++rflag;
			break;
		case 'u':	/* run with this user credentials if initially root */
			user = optarg;
			droproot++;
			break;
		case 'g':	/* run with this group credentials if initially root */
			group = optarg;
			droproot++;
			break;
		case 'D':	/* doesn't become a daemon, and stay in the foreground. */
			daemonize = 0;
			break;
		case 'm':
			save_min_record_size = (u_int)strtoul(optarg, NULL, 0);
			break;
		case 'M':
			pipe_min_record_size = (u_int)strtoul(optarg, NULL, 0);
			break;
		default:
			usage();
		}

	if (infile) 
		cmdbuf = read_infile(infile);
	else
		cmdbuf = copy_argv(&argv[optind]);

	openlog(program_name, LOG_PID|LOG_CONS, LOG_DAEMON);

	if (device_name == 0)
		if ((device_name = getenv("IFF_LISTEN")) == NULL)
			if ((device_name = pcap_lookupdev(ebuf)) == 0)
				errorx("pcap_lookupdev: %s", ebuf);

	/* Attach bpf interface to network interface */
	pd = pcap_open_live(device_name, snaplen, !pflag, 1000, ebuf);
	if (pd == NULL)
		errorx("%s: pcap_open_live: %s", device_name, ebuf);

	if (droproot) {
		struct passwd *pwd;
		struct group *grp;
		char *cp;

		if (setuid(0) != 0)
			errorx("must be run under super-user to change user id");
		if (user == NULL || *user == '\0' || group == NULL || *group == '\0')
			errorx("should specify both user and group");

		if (user[0] == '#') {
			uid = strtoul(&user[1], &cp, 10);
			if (&user[1] == cp)
				errorx("invalid uid: %s", &user[1]);
		} else {
			if ((pwd = getpwnam(user)) == NULL)
				errorx("no such user: %s", user);
			uid = pwd->pw_uid;
		}

		if (group[0] == '#') {
			gid = strtoul(&group[1], &cp, 10);
			if (&group[1] == cp)
				errorx("invalid uid: %s", &group[1]);
		} else {
			if ((grp = getgrnam(group)) == NULL)
				errorx("no such group: %s", group);
			gid = grp->gr_gid;
		}
	}

	/* chroot to TRAFDDIR */
	if (Cflag) {
		if (chdir(TRAFDDIR) == -1)
			error("cannot chdir to %s", TRAFDDIR);
		syslog(LOG_NOTICE, "(%s) chroot to %s", device_name, TRAFDDIR);
		if (chroot(TRAFDDIR) == -1)
			error("cannot chroot to %s", TRAFDDIR);
	}

	/* drop root privileges. */
	if (droproot) {
		syslog(LOG_NOTICE, "(%s) setting user to %s:%s", device_name,
			user, group);

		/* clear groups list */
		if (setgroups(0, NULL))
			error("setgroups: error clearing group list");

		if (setgid(gid))
			error("setgid");

		if (setuid(uid))
			error("setuid");
	}

	if (DevFileName == 0)
		DevFileName = device_name;

	i = pcap_snapshot(pd);
	if (snaplen < i) {
		warningx("%s: snaplen raised from %d to %d", device_name, snaplen, i);
		snaplen = i;
	}

	if (pcap_lookupnet(device_name, &localnet, &netmask, ebuf) < 0) {
		localnet = 0;
		netmask = 0;
		warningx("%s: pcap_lookupnet: %s", device_name, ebuf);
	}

	if(pcap_compile(pd, &fcode, cmdbuf, Oflag, netmask) < 0)
		errorx("%s: pcap_compile: %s", device_name, pcap_geterr(pd));

	if (dflag) {
		bpf_dump(&fcode, dflag);
		exit(0);
	}
	init_addrtoname(localnet, netmask);

	/* Initialize file names */
	if (Cflag) {
		snprintf(file_pid, sizeof(file_pid), FMT_PIDFILE, DevFileName);
		snprintf(file_out, sizeof(file_out), FMT_DUMPFILE, DevFileName);
		snprintf(file_backup, sizeof(file_backup), FMT_SAVEFILE, DevFileName);
		snprintf(file_sock, sizeof(file_sock), FMT_SOCKET, DevFileName);
	} else {
		snprintf(file_pid, sizeof(file_pid), TRAFDDIR FMT_PIDFILE,
			DevFileName);
		snprintf(file_out, sizeof(file_out), TRAFDDIR FMT_DUMPFILE,
			DevFileName);
		snprintf(file_backup, sizeof(file_backup), TRAFDDIR FMT_SAVEFILE,
			DevFileName);
		snprintf(file_sock, sizeof(file_sock), TRAFDDIR FMT_SOCKET,
			DevFileName);
	}

	/* check if pid file already exists. */
	if (access(file_pid, F_OK) == 0)
		errorx("%s exists, probably %s is already running on %s, exiting",
		      file_pid, program_name, device_name);

	/* check if socket file already exists. */
	if (access(file_sock, F_OK) == 0)
		errorx("%s exists, probably %s is already running on %s, exiting",
		      file_sock, program_name, device_name);

	/* open socket for communications */
	if ((comm_fd = socket(PF_LOCAL, SOCK_STREAM, 0)) < 0)
		error("socket PF_LOCAL SOCK_STREAM");
	sun.sun_family = AF_LOCAL;
	strlcpy(sun.sun_path, file_sock, sizeof(sun.sun_path));
	if (bind(comm_fd, (struct sockaddr *)&sun, sizeof(sun)) == -1) {
		close(comm_fd);
		error("bind %s", sun.sun_path);
	}
	if (listen(comm_fd, 1) == -1) {
		close(comm_fd);
		error("listen");
	}

	if (daemonize) {
		/* Jump to background */
		daemon(1, 0);
	}

	/* Initialize trafd stats routine */
	init_trafd_stats();
	/* Catch signals */
	sigflag_shutdown = sigflag_save = sigflag_backup = 0;
#if 0
	sigflag_info = 0;
#endif
	(void)signal(SIGTERM, sighandler_shutdown);
	(void)signal(SIGINT, sighandler_shutdown);
	(void)signal(SIGQUIT, sighandler_shutdown);
	(void)signal(SIGCHLD, sighandler_child);
	(void)signal(SIGUSR1, sighandler_save);
	(void)signal(SIGUSR2, sighandler_backup);
#if 0
	(void)signal(SIGINFO, sighandler_info);
#endif

	/* write pid file */
	if (writepid(file_pid, getpid()) == -1)
		goto errexit;

	if (pcap_setfilter(pd, &fcode) == -1) {
		(void)syslog(LOG_ERR, "(%s) traffic collector aborted, pcap error: %s",
			device_name, pcap_geterr(pd));
		pcap_close(pd);
		unlink(file_pid);
		exit(1);
	}

	/* init collector structures */
	if (traf_init(rflag) == -1)
		goto errexit;

	syslog(LOG_NOTICE, "(%s) traffic collector started", device_name);

	if (main_loop(pd, cnt, comm_fd) == -1)
		goto errexit;

	cleanup_exit();	/* clean exit */

errexit:
	/* Abnormal termination */
	(void)syslog(LOG_ERR, "(%s) traffic collector aborted: %m",
	     device_name);
	pcap_close(pd);
	unlink(file_pid);
	unlink(file_sock);
	exit(1);

	/* NOTREACHED */
	return 0;
}

static int
comm_server(int fd)
{
	int rc;

	rc = fork();
	if (rc == -1) {
		syslog(LOG_ERR, "(%s) fork: %m", device_name);
		exit(1);
	}

	if (rc) /* parent */
		return 0;

	/* child */
	rc = traf_send(fd);
	if (rc == -1)
		syslog(LOG_ERR, "(%s) write failed: %m", device_name);
	else
		syslog(LOG_INFO, "(%s) %d records sent",
			device_name, rc);
	close(fd);
	exit(0);
}

static int
process_signals()
{
#if 0
	if (sigflag_info) {	/* Write trafd internal stats to log */
		sigflag_info = 0;
		log_trafd_stats(device_name, pd);
	}
#endif
	if (sigflag_save) {	/* save and clear data */
		sigflag_save = 0;
		log_trafd_stats(device_name, pd);
		if (fork() == 0) {
			traf_save(file_backup, "a");
			exit(0);
		}
		traf_clear();
	}
	if (sigflag_backup) {	/* dump data for recovery */
		sigflag_backup = 0;
		if (fork() == 0) {
			traf_save(file_out, "w");
			exit(0);
		}
	}
	return 0;
}

static int
main_loop(pcap_t *pcap, int cnt, int comm_fd)
{
	pcap_handler printer;
	u_char *pcap_userdata;
	fd_set rfds;
	int nfds;
	int pcap_fd;
	int rc;

	printer = lookup_printer(pcap_datalink(pcap));
	pcap_userdata = NULL;

	FD_ZERO(&rfds);
	pcap_fd = pcap_fileno(pcap);
	nfds = max(pcap_fd, comm_fd) + 1;

	while (!sigflag_shutdown) {
		FD_SET(pcap_fd, &rfds);
		FD_SET(comm_fd, &rfds);
		if (select(nfds, &rfds, NULL, NULL, NULL) == -1) {
			if (errno == EINTR) {
				/* interrupted by signal */
				process_signals();
				/* restart, check for exit condition will be done
				 * at the start of a loop */
				continue;
			} else
				return -1;
		}
		process_signals();
		if (FD_ISSET(pcap_fd, &rfds)) {	/* packets available on a pcap */
			rc = pcap_dispatch(pcap, 1, printer, pcap_userdata);
			if (rc == -1) {
				syslog(LOG_ERR, "(%s) pcap_dispatch: %s", device_name,
					pcap_geterr(pcap));
				return -1;
			}
			if (cnt > 0) {
				cnt -= rc;
				if (cnt <= 0)
					return 0;	/* success */
			}
		}
		if (FD_ISSET(comm_fd, &rfds)) {	/* connection on socket */
			int fd = accept(comm_fd, NULL, NULL);
			if (fd == -1)
				return -1;
			if (comm_server(fd) == -1)
				return -1;
			close(fd);
		}
	}
	return 0;
}

/* make a clean exit */
static void
cleanup_exit()
{
	log_trafd_stats(device_name, pd);
	pcap_close(pd);
	traf_save(file_out, "w");
	(void)syslog(LOG_NOTICE, "(%s) traffic collector stopped", device_name);
	unlink(file_pid);
	unlink(file_sock);
	exit(0);
}

#define	WAIT_ACTION	5

static void
sighandler_shutdown()
{
	sigflag_shutdown++;
}

static void
sighandler_save()
{
	sigflag_save++;
}

static void
sighandler_backup()
{
	sigflag_backup++;
}

static void
sighandler_child()
{
	int status;

	/* On some systems (e.g. *BSDs) handler called if one or _more_
	 * children have died. Catch all of them. */
	for(; waitpid((pid_t)-1, &status, WNOHANG) != (pid_t)-1; );
	if (errno && errno != ECHILD && errno != EINTR)
		(void)syslog(LOG_WARNING, "(%s) waitpid: %m", device_name);
}

#if 0
static void
sighandler_info()
{
	sigflag_info++;
}
#endif

static inline int
init_trafd_stats(void)
{
	return get_trafd_stats(NULL, NULL, NULL, NULL);
}

/* Sets pointers to the static structures, containing
 * internal trafd stats. ts1 -- stats since initialization,
 * ts2 -- stats since last get_trafd_stats() call. */
static int
get_trafd_stats(char *devname, pcap_t *pcap,
	struct trafd_stats **ts1p, struct trafd_stats **ts2p)
{
	static struct trafd_stats ts0, ts1, ts2;
	static struct timeval tv0;

	if (ts1p == NULL) {
		/* Initialize and return */
		gettimeofday(&tv0, NULL);
		bzero(&ts0, sizeof(ts0));
		bcopy(&tv0, &ts0.tv, sizeof(ts0.tv));
		return 0;
	}

	if (pcap_stats(pcap, &ts1.ps) == -1) {
		syslog(LOG_WARNING, "(%s) can't obtain pcap stats: %s",
			devname, pcap_geterr(pcap));
		return -1;
	}
	gettimeofday(&ts1.tv, NULL);
	getrusage(RUSAGE_SELF, &ts1.ru);

	/* times since last info output */
	TIMEDIFF(&ts0.tv, &ts1.tv, &ts2.tv);
	TIMEDIFF(&ts0.ru.ru_utime, &ts1.ru.ru_utime, &ts2.ru.ru_utime);
	TIMEDIFF(&ts0.ru.ru_stime, &ts1.ru.ru_stime, &ts2.ru.ru_stime);

	/* packets since last info output */
	ts2.ps.ps_recv = ts1.ps.ps_recv - ts0.ps.ps_recv;
	ts2.ps.ps_drop = ts1.ps.ps_drop - ts0.ps.ps_drop;

	/* save current stats for later usage */
	bcopy(&ts1, &ts0, sizeof(ts0));

	/* uptime */
	TIMEDIFF(&tv0, &ts1.tv, &ts1.tv);

	*ts1p = &ts1;
	*ts2p = &ts2;
	return 0;
}

/* Write trafd internal statistics to log */
static int
log_trafd_stats(char *devname, pcap_t *pcap)
{
	struct trafd_stats *ts1p, *ts2p;

	if (get_trafd_stats(devname, pcap, &ts1p, &ts2p) == -1)
		return -1;

	syslog(LOG_INFO, "(%s) stats (real/user/sys/pkts/drops) "
		"since start: %ld.%02ld/%ld.%02ld/%ld.%02ld/%u/%u, "
		"since last info: %ld.%02ld/%ld.%02ld/%ld.%02ld/%u/%u",
		devname,
		ts1p->tv.tv_sec, ts1p->tv.tv_usec/10000,
		ts1p->ru.ru_utime.tv_sec, ts1p->ru.ru_utime.tv_usec/10000,
		ts1p->ru.ru_stime.tv_sec, ts1p->ru.ru_stime.tv_usec/10000,
		ts1p->ps.ps_recv, ts1p->ps.ps_drop,
		ts2p->tv.tv_sec, ts2p->tv.tv_usec/10000,
		ts2p->ru.ru_utime.tv_sec, ts2p->ru.ru_utime.tv_usec/10000,
		ts2p->ru.ru_stime.tv_sec, ts2p->ru.ru_stime.tv_usec/10000,
		ts2p->ps.ps_recv, ts2p->ps.ps_drop);

	return 0;
}

void
usage()
{
	extern char version[];

	fprintf(stderr, "trafd v%s - tcp/udp data traffic collector daemon\n",
		version);
	fprintf(stderr,
		"Usage: %s [-dOprU] [-c count] [-i iface] [-f savefile] [-F file | expr]\n"
		"\t[-u user|#uid] [-g group|#gid]\n",
		program_name);
	exit(-1);
}
