/*-
 * Copyright (c) 2006 Stanislav A. Svirid. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *  This product includes software developed by the University of
 *  California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $RISS: trafd/trafdb/main.c,v 1.1 2006/08/31 10:49:19 count Exp $
 *
 */

#include <sys/types.h>
#include <sys/stat.h>

#include <err.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sysexits.h>
#include <time.h>

#include "trafdb.h"

/*
 * Display usage information and exit
 */
static void
usage(void)
{
	fprintf(stderr, "usage: trafdb -hlO file\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "    -h\tthis help message\n");
	fprintf(stderr, "    -c\tjust copy entries to output\n");
	fprintf(stderr, "    -l\tlist entries\n");
	fprintf(stderr, "    -b\tbegin date or entry number (use #)\n");
	fprintf(stderr, "    -e\tend date or entry number (use #)\n");
	fprintf(stderr, "    -D FMT\tUse user-defined date format\n");
	fprintf(stderr, "    -A\tappend output file\n");
	fprintf(stderr, "    -C\tcompatible with old-style db\n");
	fprintf(stderr, "    -S\tsummary output\n");
	fprintf(stderr, "    -R\treplace begin date with end date\n");
	fprintf(stderr, "\n");

	exit(EX_USAGE);
}


static char date_format[13] = "%y%m%d%H%M";

static void
setthetime(const char *_str, time_t *_time)
{
	struct tm   *tm;
	u_long	    num;
	time_t	    now;
	int	    offset;

	now = time(NULL);
	tm = localtime(&now);
	tm->tm_isdst = -1; /* we don't need it */
	tm->tm_sec = 0;

	if (_str == NULL)
	    return;

	if (*_str == '#') {
	    num = strtoul(_str + 1, NULL, 10);
	    if (num == 0)
		err(1, "Error converting '%s' to number", _str);
	    *_time = -((time_t)num);
	} else {
	    offset = strlen(date_format) - strlen(_str);

	    if (offset < 0)
		errx(1, "Length of date string '%s' too long", _str);

	    if ((offset % 2) != 0)
		errx(1, "Date string must have odd numbers");

	    if (strptime(_str, date_format + offset, tm) == NULL) {
		errx(1, "Error converting date '%s' with format '%s'",
			_str, date_format + offset);
		/* NOTREACHED */
	    }

	    *_time = mktime(tm);
	}
}

/*
 * This programm for manipulate trafd dadtabase files
 */
int
main(int ac, char *av[])
{
	struct trafdb	db;
	int	cmd;
	char	ch, *in, *out;

	cmd = TRAFDB_CMD_UNKNOWN;
	db.flags = 0;
	db.begin = 0;
	db.end = 0;

	if (ac < 3) {
	    usage();
	    /* NOTREACHED */
	}

	while((ch = getopt(ac, av, "cb:e:hlt:ACD:RS")) != -1)
	    switch (ch) {
	    case 'c':
		if (cmd != TRAFDB_CMD_UNKNOWN) {
		    warnx("Can't use multimple commands");
		    usage();
		    /* NOTREACHED */
		}
		cmd = TRAFDB_CMD_COPY;
		break;
	    case 'b':
		setthetime(optarg, &db.begin);
		break;
	    case 'e':
		setthetime(optarg, &db.end);
		break;
	    case 'l':
		if (cmd != TRAFDB_CMD_UNKNOWN) {
		    warnx("Can't use multimple commands");
		    usage();
		    /* NOTREACHED */
		}
		cmd = TRAFDB_CMD_LIST;
	        break;
	    case 'A':
		db.flags |= TRAFDB_APPEND;
		break;
	    case 'C':
		db.flags |= TRAFDB_COMPAT;
		break;
	    case 'D':
		strncpy(date_format, optarg, sizeof(date_format));
		break;
	    case 'R':
		cmd = TRAFDB_CMD_REPLACE;
		break;
	    case 'S':
		db.flags |= TRAFDB_SUMMARY;
		break;
	    case 'h':
	    default:
	        usage();
	        /* NOTREACHED */
	    }
	ac -= optind;
	av += optind;

	if (ac == 0) {
	    usage();
	    /* NOTREACHED */
	}

	in = av[0];
	out = ac > 1 ? av[1] : NULL;

/*
	printf("begin: ");
	if (db.begin > 0)
	    printf("%s\n", ctime(&db.begin));
	else
	    printf("#%d\n", db.begin);

	printf("end: ");
	if (db.end > 0)
	    printf("%s\n", ctime(&db.end));
	else
	    printf("#%d\n", db.end);
*/
	trafdb_init(in, out, &db);

	if (cmd == TRAFDB_CMD_LIST) {
	    trafdb_list_entries(&db);
	} else if (cmd == TRAFDB_CMD_COPY) {
	    trafdb_copy_entries(&db);
	} else if (cmd == TRAFDB_CMD_REPLACE) {
	    trafdb_replace(&db);
	}

	trafdb_close(&db);

	exit(EX_OK);
}
