/*-
 * Copyright (c) 2006 Stanislav A. Svirid. All Rights Reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *  This product includes software developed by the University of
 *  California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $RISS$
 *
 */

#include <sys/types.h>
#include <sys/stat.h>

#include <err.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sysexits.h>
#include <unistd.h>

#include "trafdb.h"

/*
 * Initialize traf db
 */
void
trafdb_init(const char *_in_name, const char *_out_name, trafdb_t _db)
{
	FILE *in, *out;
	size_t max_str_len;

	max_str_len = sizeof(_db->in_name) - 1;

	if (_in_name == NULL || strcmp(_in_name, "-") == 0) {
	    in = stdin;
	    strncpy(_db->in_name, "(stdin)", max_str_len);
	}
	else {
	    in = fopen(_in_name, "r");
	    strncpy(_db->in_name, _in_name, max_str_len);
	}
	_db->in_name[max_str_len] = '\0';

	if (in == NULL)
	    err(1, "Can't open database file %s for read", _db->in_name);


	if (_out_name == NULL || strcmp(_out_name, "-") == 0) {
	    out = stdout;
	    strncpy(_db->out_name, "(stdout)", max_str_len);
	}
	else {
	    out = fopen(_out_name, is_append(_db) ? "a" : "w");
	    strncpy(_db->out_name, _out_name, max_str_len);
	}
	_db->out_name[max_str_len] = '\0';

	if (out == NULL)
	    err(1, "Can't open database file %s for write", _db->out_name);


	_db->_in = in;
	_db->_out = out;
	_db->entry = 0;
	_db->last.data = NULL;

	if (trafdb_get_header(_db) != 0)
	    err(1, "Error reading first header");
}

/*
 * Cleanups and close db file
 */
void
trafdb_close(trafdb_t _db)
{
	if (_db->last.data != NULL)
	    free(_db->last.data);

	fclose(_db->_in);
	fclose(_db->_out);
}

/*
 * Check last entry in properly range
 */
int
trafdb_in_range(trafdb_t _db)
{
    tdb_entry_t	e = &_db->last;
    int	ret = 0;

    if (_db->begin != 0) {
	if (_db->begin > 0) {
	    if (_db->begin > e->begin)
		ret++;
	} else {
	    if (-(_db->begin) > _db->entry)
		ret++;
	}
    }

    if (_db->end != 0) {
	if (_db->end > 0) {
	    if (_db->end < e->end)
		ret++;
	} else {
	    if (-(_db->end) < _db->entry)
		ret++;
	}
    }

    return ret;
}

/*
 * Read old trafd db header
 * Try found first header in range
 */
static int
compat_get_header(trafdb_t _db)
{
	struct timeval	begin, end;
	ssize_t	len;

	len = fread(&_db->last.records, 1, sizeof(_db->last.records),
		_db->_in);
	if (len != sizeof(_db->last.records)) {
	    if (feof(_db->_in))
		return 1; /* eof reached */

	    err(1, "Error reading header");
	}

	len = fread(&begin, 1, sizeof(begin), _db->_in);
	if (len != sizeof(begin))
	    err(1, "Error reading header");

	len = fread(&end, 1, sizeof(end), _db->_in);
	if (len != sizeof(end))
	    err(1, "Error reading header");

	_db->last.begin = begin.tv_sec;
	_db->last.end = end.tv_sec;
	_db->entry++;
	_db->last.record_size = sizeof(struct old_record);

	return 0;
}

/*
 * Read new traf db header
 * Try found first header in range
 */
static int
new_get_header(trafdb_t _db)
{
	trafdb_close(_db);
	errx(1, "New DB version is not supported yet");
	/* NOTREACHED */

	return 1;
}

/*
 * Read entry header
 * Try found first header in range
 */
int
trafdb_get_header(trafdb_t _db)
{
	return is_compat(_db) ? compat_get_header(_db) : new_get_header(_db);
}

/*
 * Skip all records
 */
void
trafdb_skip_records(trafdb_t _db)
{
	int ret;

	if (_db->last.records > 0) {
	    if (_db->_in != stdin) {
		ret = fseeko(_db->_in,
		    _db->last.record_size * _db->last.records,
		    SEEK_CUR);
		if (ret == -1)
		    err(1, "Error in seek");
	    } else {
		trafdb_load_records(_db);
	    }
	}
}

/*
 * Load records
 */
void
trafdb_load_records(trafdb_t _db) {
	tdb_entry_t	h = &_db->last;
	int	len;

        if (h->data != NULL)
	    free(h->data);

	h->data = malloc(h->records * h->record_size);
	if (h->data == NULL) {
	    trafdb_close(_db);
	    err(1, "Can't allocate %d bytes", h->records * h->record_size);
	    /* NOTREACHED */
	}

	len = fread(h->data, h->record_size, h->records, _db->_in);
	if (len != h->records) {
	    trafdb_close(_db);
	    err(1, "Error reading records");
	    /* NOTREACHED */
	}
}

/*
 * Get specified record
 */
tdb_record_t
trafdb_get_record(trafdb_t _db, int rec) {
	if (_db->last.data == NULL || rec >= _db->last.records || rec < 0)
	    return NULL;

	return (tdb_record_t)(_db->last.data + rec * _db->last.record_size);
}

/*
 * List entries from OLD db
 */
static void
compat_list_entries(trafdb_t _db)
{
	tdb_record	sum, *temp;
	int	i, range;
	char	buf[16];

	bzero(&sum, sizeof(tdb_record));

	do {
	    range = trafdb_in_range(_db);
	    if (range == 2)
		break;
	    if (range != 0) {
		trafdb_skip_records(_db);
		continue;
	    }

	    strftime(buf, 15, "%b %e %T", localtime(&_db->last.begin));
	    buf[15] = '\0';
	    fprintf(_db->_out, "%03u   %.15s", _db->entry, buf);

	    strftime(buf, 15, "%b %e %T", localtime(&_db->last.end));
	    buf[15] = '\0';
	    fprintf(_db->_out, "   %.15s", buf);

	    if (is_sum(_db)) {
		trafdb_load_records(_db);

		for(i = 0; i < _db->last.records; i++) {
		    temp = trafdb_get_record(_db, i);
		    if (temp == NULL)
			errx(1, "Can't get record %d", i);

		    sum.old.n_psize += temp->old.n_psize;
		    sum.old.n_bytes += temp->old.n_bytes;
		}

		fprintf(_db->_out, "   %10u   %10u",
			sum.old.n_bytes, sum.old.n_psize);
	    } else {
		trafdb_skip_records(_db);
	    }

	    fprintf(_db->_out, "   %6d\n", _db->last.records);
	} while (trafdb_get_header(_db) == 0);
}

/*
 * List entries from NEW db
 */
static void
new_list_entries(trafdb_t _db)
{
    return;
}

/*
 * List entries from db
 */
void
trafdb_list_entries(trafdb_t _db)
{
    is_compat(_db) ? compat_list_entries(_db) : new_list_entries(_db);
}

/*
 * Copy out OLD entries
 */
static int
compat_save_header(trafdb_t _db, tdb_entry_t _entry)
{
	struct timeval	begin, end;
	int len;

	begin.tv_sec = _entry->begin;
	end.tv_sec = _entry->end;

	len = fwrite(&_db->last.records, 1, sizeof(_db->last.records),
		_db->_out);
	if (len != sizeof(_db->last.records))
		return 1;

	len = fwrite(&begin, 1, sizeof(begin), _db->_out);
	if (len != sizeof(begin))
	    return 1;

	len = fwrite(&end, 1, sizeof(end), _db->_out);
	if (len != sizeof(end))
	    return 1;

	return 0;
}

/*
 * Copy out NEW entries
 */

static int
new_save_header(trafdb_t _db, tdb_entry_t _entry)
{
    return 1;
}

/*
 * Save one entry to stream
 */

void
trafdb_save_entry(trafdb_t _db, tdb_entry_t _entry)
{
	int ret;

	ret = is_compat(_db) ?
		compat_save_header(_db, _entry) :
		new_save_header(_db, _entry);

	if (ret != 0)
	    err(1, "Error saving header for entry '%d'", _db->entry);

	if (_entry->records) {
	    ret = fwrite(_entry->data, _entry->record_size,
		    _entry->records, _db->_out);

	    if (ret != _entry->records)
		err(1, "Error saving records for entry '%d'", _db->entry);
	}
}

/*
 * Copy out entries
 */
void
trafdb_copy_entries(trafdb_t _db)
{
	int range = 0;

	do {
	    range = trafdb_in_range(_db);
	    if (range == 2)
		break;
	    if (range != 0) {
		trafdb_skip_records(_db);
		continue;
	    }

	    trafdb_load_records(_db);
	    trafdb_save_entry(_db, &_db->last);
	} while (trafdb_get_header(_db) == 0);
}

/*
 * Replace time record 'begin' with 'end'
 */
void
trafdb_replace(trafdb_t _db)
{
    tdb_entry_t	e = &_db->last;

    if (_db->begin == 0 || _db->end == 0)
	errx(1, "You must specify both begin and end date");

    do {
	if (e->begin == _db->begin)
	    e->begin = _db->end;

	if (e->end == _db->begin)
	    e->end = _db->end;

	trafdb_load_records(_db);
	trafdb_save_entry(_db, &_db->last);
    } while (trafdb_get_header(_db) == 0);
}
